/* vim: set foldmethod=marker : */
#include <avr/io.h>
#include <avr/interrupt.h>
#include <util/delay.h>
#include <stdint.h>
#include <stdio.h>

#define UCSC_MASK (_BV(UMSEL) | (2 << UCSZ0))

#define SPI_PORT PORTB
#define SPI_PIN  PINB
#define SPI_DDR  DDRB

#define SPI_BIT_CS 1
#define SPI_BIT_SS 2
#define SPI_BIT_MOSI 3
#define SPI_BIT_MISO 4
#define SPI_BIT_SCK 5

/* Скорость обмена по UART */
#define UART_BITRATE 19200UL

#define UBRR_VALUE(f, br) (f / br / 16 - 1)

static int uart_putchar(char c, FILE *stream);
static FILE mystdout = FDEV_SETUP_STREAM(uart_putchar, NULL,
										 _FDEV_SETUP_WRITE);
static void uart_tx(uint8_t c)
{
	while (!(UCSRA & _BV(UDRE)));
	UDR = c;
}

static int uart_putchar(char c, FILE *stream)
{
	if (c == '\n') {
		uart_putchar('\r', stream);
	}
	uart_tx(c);
	return 0;
}

static void init_uart(uint16_t ubrr)/*{{{*/
{
	UCSRB = _BV(TXEN) |_BV(RXEN);
	UCSRC = UCSC_MASK;
	UBRRH = ubrr >> 8;
	UBRRL = ubrr & 0xFF;
}/*}}}*/


uint16_t cnth = 0;
int main(void)
{

	/* Конфигурация SPI */
	SPI_DDR = _BV(SPI_BIT_SS) | _BV(SPI_BIT_CS) | _BV(SPI_BIT_MOSI) | _BV(SPI_BIT_SCK);
	/*DDRB = 0xB0;*/

	SPI_PORT |= _BV(SPI_BIT_CS) | _BV(SPI_BIT_MISO) | _BV(SPI_BIT_SS);
	/* PORTB |= _BV(6) | _BV(4); */

	SPCR = _BV(SPE) | _BV(MSTR);
	_delay_ms(10);
	_delay_ms(10);

	init_uart(UBRR_VALUE(F_CPU, UART_BITRATE));

	stdout = &mystdout;
	sei();

	SPI_PORT &= ~(_BV(SPI_BIT_CS) | _BV(SPI_BIT_SS));
	_delay_us(1);
	SPI_PORT |= (_BV(SPI_BIT_CS) | _BV(SPI_BIT_SS));
	_delay_us(40);
	SPI_PORT &= ~(_BV(SPI_BIT_CS) | _BV(SPI_BIT_SS));
	SPDR = 0x30;
	while (!(SPSR & _BV(SPIF)));
	SPI_PORT |= (_BV(SPI_BIT_CS) | _BV(SPI_BIT_SS));

	_delay_ms(10);

	SPI_PORT &= ~(_BV(SPI_BIT_CS) | _BV(SPI_BIT_SS));
	while (SPI_PIN & _BV(SPI_BIT_MISO));
	SPDR = 2;
	while (!(SPSR & _BV(SPIF)));
	SPDR = 0x2E;
	while (!(SPSR & _BV(SPIF)));
	SPI_PORT |= (_BV(SPI_BIT_CS) | _BV(SPI_BIT_SS));

	SPI_PORT &= ~(_BV(SPI_BIT_CS) | _BV(SPI_BIT_SS));
	while (SPI_PIN & _BV(SPI_BIT_MISO));
	SPDR = 0;
	while (!(SPSR & _BV(SPIF)));
	SPDR = 0x34;
	while (!(SPSR & _BV(SPIF)));
	SPI_PORT |= (_BV(SPI_BIT_CS) | _BV(SPI_BIT_SS));
	ASSR |= _BV(AS2);
	ADMUX = _BV(REFS0) | (0 << MUX0);
	uint16_t i = 1;
	while (1) {
		_delay_ms(40);
		ADCSRA = _BV(ADEN) | _BV(ADSC) | _BV(ADIF) | (5 << ADPS0);
		while (!(ADCSRA & _BV(ADIF)));
		uint8_t adcl = ADCL;
		uint8_t adch = ADCH;
		uint32_t adc = (adch << 8) | adcl;
		uint32_t br = (7000000UL + (adc * 1000UL)) / 16 / UART_BITRATE - 1;
		UBRRH = br >> 8;
		UBRRL = br & 0xFF;
		printf("0x%04X 0x%04X 0x%04X\n", (uint16_t)adc, (uint16_t)br, i++);
		if (UCSRA & _BV(RXC)) {
			break;
		}
	}


	while (1) {
		TCNT1 = 0;
		cnth = 0;
		TIFR |= _BV(TOV2);
		TIFR |= _BV(TOV1);
		TIMSK |= _BV(TOIE1);
		TCNT2 = 252;
		TCCR2 = 7 << CS20;
		while  (!(TIFR & _BV(TOV2)));
		TCCR1B |= 7 << CS10;
		TIFR |= _BV(TOV2);
		while  (!(TIFR & _BV(TOV2)));
		TCCR1B = 0;
		TCCR2 = 0;
		TIMSK &= ~_BV(TOIE1);
		if (TIFR & _BV(TOV1)) {
			cnth ++;
		}
		uint32_t freq = ((uint32_t)cnth << 16) + TCNT1;
		freq >>= 1;

		uint64_t FREQ = ((uint64_t)2403000000UL << 16) / freq;

		char fs[10] ;
		printf("Fxosc = ");
		uint8_t i  = 8;
		fs[9] = 0;
		while (freq) {
			fs[i] = '0' + freq %10;
			freq /= 10;
			i--;
		}
		while (i != 0xFF) {
			fs[i] = ' ';
			i--;
		}
		printf("%s\n", fs);

		printf("FREQ2 = 0x%02X\n", (uint8_t)((FREQ >> 16)& 0xFF));
		printf("FREQ1 = 0x%02X\n", (uint8_t)((FREQ >> 8)& 0xFF));
		printf("FREQ0 = 0x%02X\n", (uint8_t)((FREQ )& 0xFF));
	}
}

ISR(SIG_OVERFLOW1)
{
	cnth++;
}

